/******************************************************************************/
/****         Firebird Volltextindex                                       ****/
/******************************************************************************/



/******************************************************************************/
/****                               Domains                                ****/
/******************************************************************************/

CREATE DOMAIN "dom_BOOL" AS
SMALLINT
NOT NULL
CHECK (VALUE IN (0, 1));

CREATE DOMAIN "dom_Hash" AS
BIGINT
NOT NULL;

CREATE DOMAIN "dom_PK" AS
INTEGER
NOT NULL;

CREATE DOMAIN "dom_Position" AS
INTEGER
NOT NULL;

CREATE DOMAIN "dom_RDB$NAME" AS
VARCHAR(31) CHARACTER SET UNICODE_FSS;

CREATE DOMAIN "dom_SingleChar" AS
CHAR(1) CHARACTER SET UTF8;

CREATE DOMAIN "dom_TS" AS
TIMESTAMP;

CREATE DOMAIN "dom_Text" AS
BLOB SUB_TYPE 1 SEGMENT SIZE 80 CHARACTER SET UTF8;

CREATE DOMAIN "dom_TextCache" AS
VARCHAR(8191) CHARACTER SET UTF8;

CREATE DOMAIN "dom_Word" AS
VARCHAR(1021) CHARACTER SET UTF8
NOT NULL;

CREATE DOMAIN "dom_WordLength" AS
SMALLINT;

CREATE DOMAIN "dom_WordType" AS
SMALLINT
NOT NULL;



/******************************************************************************/
/****                              Generators                              ****/
/******************************************************************************/

CREATE GENERATOR "gen_PK";


SET TERM ^ ; 



/******************************************************************************/
/****                          Stored Procedures                           ****/
/******************************************************************************/

CREATE OR ALTER PROCEDURE "prc_domain_text_length" (
    IN_DOM_NAME "dom_RDB$NAME" COLLATE UNICODE_FSS)
RETURNS (
    OUT_TEXT_LENGTH "dom_WordLength")
AS
BEGIN
  SUSPEND;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_index_char" (
    IN_TEXT TYPE OF "dom_TextCache" COLLATE UTF8,
    IN_WHOLE_WORD TYPE OF "dom_BOOL",
    IN_WORD_BEGINNINGS TYPE OF "dom_BOOL",
    IN_WORD_TRAILINGS TYPE OF "dom_BOOL",
    IN_WORD_PARTS TYPE OF "dom_BOOL",
    IN_MIN_WORD_PART_LENGTH TYPE OF "dom_WordLength" = 1,
    IN_POSITION_OFFSET TYPE OF "dom_Position" = 0,
    IN_WORD_INDEX_OFFSET TYPE OF "dom_Position" = 0)
RETURNS (
    OUT_WORD_INDEX INTEGER)
AS
BEGIN
  SUSPEND;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_index_text" (
    IN_TEXT TYPE OF "dom_Text",
    IN_WHOLE_WORD TYPE OF "dom_BOOL" = 1,
    IN_WORD_BEGINNINGS TYPE OF "dom_BOOL" = 1,
    IN_WORD_TRAILINGS TYPE OF "dom_BOOL" = 1,
    IN_WORD_PARTS TYPE OF "dom_BOOL" = 1,
    IN_MIN_WORD_PART_LENGTH TYPE OF "dom_WordLength" = 1,
    IN_POSITION_OFFSET TYPE OF "dom_Position" = 0,
    IN_WORD_INDEX_OFFSET TYPE OF "dom_Position" = 0)
RETURNS (
    OUT_WORT_INDEX INTEGER)
AS
BEGIN
  SUSPEND;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_is_alpha_num" (
    IN_CHAR "dom_SingleChar" COLLATE UTF8)
RETURNS (
    OUT_IS_ALPHANUM "dom_BOOL")
AS
BEGIN
  SUSPEND;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_word_index" (
    IN_WORD "dom_Word" COLLATE UTF8,
    IN_WORD_START "dom_Position",
    IN_WORD_INDEX "dom_Position",
    IN_WHOLE_WORD "dom_BOOL",
    IN_WORD_BEGINNINGS "dom_BOOL",
    IN_WORD_TRAILINGS "dom_BOOL",
    IN_WORD_PARTS "dom_BOOL",
    IN_MIN_WORD_PART_LENGTH "dom_WordLength" = 1)
RETURNS (
    OUT_WORD_INDEX "dom_Position")
AS
BEGIN
  SUSPEND;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_word_link" (
    IN_WORD "dom_Word" COLLATE UTF8,
    IN_WORD_INDEX "dom_Position",
    IN_WORD_POSITION "dom_Position",
    IN_WORD_TYPE "dom_WordType",
    IN_MIN_WORD_PART_LENGTH "dom_WordLength" = 1)
AS
BEGIN
  EXIT;
END^



SET TERM ; ^



/******************************************************************************/
/****                                Tables                                ****/
/******************************************************************************/



CREATE GLOBAL TEMPORARY TABLE "_temp_text_link_words" (
    "FK_ttw_wrd"       "dom_PK",
    "FK_ttw_wty"       "dom_WordType",
    "CL_ttw_index"     "dom_Position",
    "CL_ttw_position"  "dom_Position"
) ON COMMIT DELETE ROWS;


CREATE TABLE "text_link_words" (
    "FK_tlw_txt"       "dom_PK",
    "FK_tlw_wrd"       "dom_PK",
    "FK_tlw_wty"       "dom_WordType",
    "CL_tlw_index"     "dom_Position",
    "CL_tlw_position"  "dom_Position"
);


CREATE TABLE "texts" (
    "PK_txt"               "dom_PK",
    "CL_txt_text"          "dom_Text",
    "CL_txt_length"        "dom_Position",
    "CL_txt_bytes"         "dom_Position",
    "CL_txt_hash"          "dom_Hash",
    "CL_txt_update_index"  "dom_BOOL" DEFAULT 0,
    "CL_txt_update_delay"  "dom_BOOL" DEFAULT 0,
    "CL_txt_inserted"      "dom_TS" DEFAULT 'NOW' NOT NULL,
    "CL_txt_updated"       "dom_TS"
);


CREATE TABLE "word_types" (
    "PK_wty"       "dom_WordType",
    "CL_wty_name"  "dom_RDB$NAME"
);


CREATE TABLE "words" (
    "PK_wrd"       "dom_PK",
    "CL_wrd_word"  "dom_Word"
);


INSERT INTO "word_types" ("PK_wty", "CL_wty_name") VALUES (0, 'ganzes Wort');
REINSERT (1, 'Wortanfang');
REINSERT (2, 'Wortteil');
REINSERT (3, 'Wortende');

COMMIT WORK;



/******************************************************************************/
/****                             Primary Keys                             ****/
/******************************************************************************/

ALTER TABLE "texts" ADD CONSTRAINT "PK_texts" PRIMARY KEY ("PK_txt");
ALTER TABLE "word_types" ADD CONSTRAINT "PK_word_types" PRIMARY KEY ("PK_wty");
ALTER TABLE "words" ADD CONSTRAINT "PK_words" PRIMARY KEY ("PK_wrd");


/******************************************************************************/
/****                             Foreign Keys                             ****/
/******************************************************************************/

ALTER TABLE "text_link_words" ADD CONSTRAINT "FK_tlw_txt" FOREIGN KEY ("FK_tlw_txt") REFERENCES "texts" ("PK_txt") ON DELETE CASCADE ON UPDATE CASCADE;
ALTER TABLE "text_link_words" ADD CONSTRAINT "FK_tlw_wrd" FOREIGN KEY ("FK_tlw_wrd") REFERENCES "words" ("PK_wrd") ON DELETE CASCADE ON UPDATE CASCADE;
ALTER TABLE "text_link_words" ADD CONSTRAINT "FK_tlw_wty" FOREIGN KEY ("FK_tlw_wty") REFERENCES "word_types" ("PK_wty") ON DELETE CASCADE ON UPDATE CASCADE;


/******************************************************************************/
/****                               Indices                                ****/
/******************************************************************************/

CREATE UNIQUE INDEX "IDX_wrd_word" ON "words" COMPUTED BY (LOWER("CL_wrd_word"));


/******************************************************************************/
/****                               Triggers                               ****/
/******************************************************************************/


SET TERM ^ ;



/******************************************************************************/
/****                         Triggers for tables                          ****/
/******************************************************************************/



/* Trigger: "texts_AIU0" */
CREATE OR ALTER TRIGGER "texts_AIU0" FOR "texts"
ACTIVE AFTER INSERT OR UPDATE POSITION 0
AS

  DECLARE VARIABLE BLOCK_SIZE TYPE OF "dom_Position";
  DECLARE VARIABLE ID         TYPE OF "dom_PK";
  DECLARE VARIABLE WORD_INDEX TYPE OF "dom_Position";

BEGIN

  IF (
       (NEW."CL_txt_update_index" = 1)
       AND
       (NEW."CL_txt_update_delay" = 0)
     )
  THEN 
  BEGIN

    /* Text locken, damit kein anderer Prozess daran arbeitet. */
         SELECT "PK_txt"
           FROM "texts"
          WHERE "PK_txt" = NEW."PK_txt"
     FOR UPDATE OF "CL_txt_text"
      WITH LOCK
           INTO :ID;

    /* Blockgre bestimmen */
    EXECUTE PROCEDURE "prc_domain_text_length"('dom_TextCache')
      RETURNING_VALUES BLOCK_SIZE;

    /* Aufruf je nach Textlnge */
    IF (CHAR_LENGTH(NEW."CL_txt_text") <= BLOCK_SIZE)
    THEN
      EXECUTE PROCEDURE "prc_ftx_index_char"(NEW."CL_txt_text", 1, 1, 1, 1, 4)
        RETURNING_VALUES WORD_INDEX;
    ELSE
      EXECUTE PROCEDURE "prc_ftx_index_text"(NEW."CL_txt_text", 1, 1, 1, 1, 4)
        RETURNING_VALUES WORD_INDEX;

    /* alte Verknpfungen lschen */
    IF (UPDATING)
    THEN
      DELETE FROM
        "text_link_words"
       WHERE
         "FK_tlw_txt" = :ID;

    /* temporre Daten kopieren */
    INSERT INTO
      "text_link_words"
      ("FK_tlw_txt", "FK_tlw_wrd", "CL_tlw_index", "CL_tlw_position", "FK_tlw_wty")
    SELECT
      :ID, "FK_ttw_wrd", "CL_ttw_index", "CL_ttw_position", "FK_ttw_wty"
    FROM
      "_temp_text_link_words";

    /* temporre Daten lschen */
    DELETE FROM
      "_temp_text_link_words";

    /* Update-Flag zurcksetzen */
    UPDATE
      "texts"
    SET
      "CL_txt_update_index" = 0
    WHERE
      "PK_txt" = :ID;

  END

  WHEN ANY
  DO
  BEGIN
    /* temporre Daten lschen */
    DELETE FROM
      "_temp_text_link_words";
  END

END
^


/* Trigger: "texts_BIU0" */
CREATE OR ALTER TRIGGER "texts_BIU0" FOR "texts"
ACTIVE BEFORE INSERT OR UPDATE POSITION 0
AS
DECLARE VARIABLE TEXT_BUF TYPE OF "dom_Text";
BEGIN

  /* PK */
  IF (NEW."PK_txt" IS NULL)
  THEN
    NEW."PK_txt" = GEN_ID("gen_PK", 1);

  /* Zeitpunkt */
  IF (INSERTING)
  THEN 
  BEGIN
    NEW."CL_txt_inserted" = 'NOW';
  END
  ELSE
  BEGIN
    IF (UPDATING)
    THEN
    BEGIN
      NEW."CL_txt_inserted" = OLD."CL_txt_inserted";
      NEW."CL_txt_updated"  = 'NOW';
    END
  END

  /* Texteigenschaften */
  TEXT_BUF = COALESCE(NEW."CL_txt_text", '');
  NEW."CL_txt_length" = CHAR_LENGTH (TEXT_BUF);
  NEW."CL_txt_bytes"  = OCTET_LENGTH(TEXT_BUF);
  NEW."CL_txt_hash"   = HASH        (TEXT_BUF);

  /* Flag zur Berechnung des Volltextindex */
  IF (COALESCE(NEW."CL_txt_update_index", 0) = 0)
  THEN
  BEGIN
    IF (
         (NEW."CL_txt_length" <> COALESCE(OLD."CL_txt_length", 0))
         OR
         (NEW."CL_txt_bytes"  <> COALESCE(OLD."CL_txt_bytes" , 0))
         OR
         (NEW."CL_txt_hash"   <> COALESCE(OLD."CL_txt_hash"  , 0))
       )
    THEN
      NEW."CL_txt_update_index" = 1;
    ELSE
      NEW."CL_txt_update_index" = 0;
  END

END
^


/* Trigger: "words_BIU0" */
CREATE OR ALTER TRIGGER "words_BIU0" FOR "words"
ACTIVE BEFORE INSERT OR UPDATE POSITION 0
AS
BEGIN

  /* PK mu sein */
  IF (NEW."PK_wrd" IS NULL)
  THEN
    NEW."PK_wrd" = GEN_ID("gen_PK", 1);

END
^


SET TERM ; ^



/******************************************************************************/
/****                          Stored Procedures                           ****/
/******************************************************************************/


SET TERM ^ ;

CREATE OR ALTER PROCEDURE "prc_domain_text_length" (
    IN_DOM_NAME "dom_RDB$NAME" COLLATE UNICODE_FSS)
RETURNS (
    OUT_TEXT_LENGTH "dom_WordLength")
AS
BEGIN
  /* Zeichenlnge der Domain feststellen */
  SELECT
    RDB$CHARACTER_LENGTH
  FROM
    RDB$FIELDS
  WHERE
    RDB$FIELD_NAME = :IN_DOM_NAME
  INTO
    :OUT_TEXT_LENGTH;
END^


CREATE OR ALTER PROCEDURE "prc_ftx_index_char" (
    IN_TEXT TYPE OF "dom_TextCache" COLLATE UTF8,
    IN_WHOLE_WORD TYPE OF "dom_BOOL",
    IN_WORD_BEGINNINGS TYPE OF "dom_BOOL",
    IN_WORD_TRAILINGS TYPE OF "dom_BOOL",
    IN_WORD_PARTS TYPE OF "dom_BOOL",
    IN_MIN_WORD_PART_LENGTH TYPE OF "dom_WordLength" = 1,
    IN_POSITION_OFFSET TYPE OF "dom_Position" = 0,
    IN_WORD_INDEX_OFFSET TYPE OF "dom_Position" = 0)
RETURNS (
    OUT_WORD_INDEX INTEGER)
AS
DECLARE VARIABLE TEXT_LEN TYPE OF "dom_Position"; /* Textlnge */
DECLARE VARIABLE WORD_INDEX TYPE OF "dom_Position"; /* aktueller Wortindex */
DECLARE VARIABLE WORD_START TYPE OF "dom_Position"; /* Wortstartposition im Text */
DECLARE VARIABLE WORD_STOP TYPE OF "dom_Position"; /* Wortendeposition im Text */
DECLARE VARIABLE CHAR_IS_ALPHANUM TYPE OF "dom_BOOL"; /* aktuell untersuchte Position ist alphanumerisch oder nicht */
BEGIN

  /* Wortindex initialisieren */
  WORD_INDEX = IN_WORD_INDEX_OFFSET;

  /* Textlnge bestimmen */
  TEXT_LEN = COALESCE(CHAR_LENGTH(IN_TEXT), 0);
  IF (TEXT_LEN > 0)
  THEN 
  BEGIN

    WORD_START = 1;
    WHILE (WORD_START <= TEXT_LEN)
    DO
    BEGIN

      /* Wortanfang suchen */
      EXECUTE PROCEDURE "prc_ftx_is_alpha_num"(SUBSTRING(IN_TEXT FROM WORD_START FOR 1))
        RETURNING_VALUES CHAR_IS_ALPHANUM;
      WHILE (
              (WORD_START <= TEXT_LEN)
              AND
              (CHAR_IS_ALPHANUM = 0)
            )
      DO
      BEGIN
        WORD_START = WORD_START + 1;
        IF (WORD_START <= TEXT_LEN)
        THEN
          EXECUTE PROCEDURE "prc_ftx_is_alpha_num"(SUBSTRING(IN_TEXT FROM WORD_START FOR 1))
            RETURNING_VALUES CHAR_IS_ALPHANUM;
      END
      /* Wortanfang gefunden? */
      IF (CHAR_IS_ALPHANUM = 0)
      THEN
        BREAK;
      
      /* Wortende suchen */
      WORD_STOP = WORD_START;
      WHILE (
              (WORD_STOP <= TEXT_LEN)
              AND
              (CHAR_IS_ALPHANUM = 1)
            )
      DO
      BEGIN
        WORD_STOP = WORD_STOP + 1;
        IF (WORD_STOP <= TEXT_LEN)
        THEN
          EXECUTE PROCEDURE "prc_ftx_is_alpha_num"(SUBSTRING(IN_TEXT FROM WORD_STOP FOR 1))
            RETURNING_VALUES CHAR_IS_ALPHANUM;
      END
      
      IF (WORD_STOP > WORD_START)
      THEN
        /* Wort extrahieren und indizieren */
        EXECUTE PROCEDURE "prc_ftx_word_index"(SUBSTRING(IN_TEXT FROM WORD_START FOR WORD_STOP - WORD_START), WORD_START + IN_POSITION_OFFSET, WORD_INDEX, IN_WHOLE_WORD, IN_WORD_BEGINNINGS, IN_WORD_TRAILINGS, IN_WORD_PARTS, IN_MIN_WORD_PART_LENGTH)
          RETURNING_VALUES WORD_INDEX;

      /* nchstes Wort */
      WORD_START = WORD_STOP + 1;

    END /* WHILE (WORD_START <= TEXT_LEN) */

  END /* IF (TEXT_LEN > 0) */

  /* Wortindex zurckgeben */
  OUT_WORD_INDEX = WORD_INDEX;

END^


CREATE OR ALTER PROCEDURE "prc_ftx_index_text" (
    IN_TEXT TYPE OF "dom_Text",
    IN_WHOLE_WORD TYPE OF "dom_BOOL" = 1,
    IN_WORD_BEGINNINGS TYPE OF "dom_BOOL" = 1,
    IN_WORD_TRAILINGS TYPE OF "dom_BOOL" = 1,
    IN_WORD_PARTS TYPE OF "dom_BOOL" = 1,
    IN_MIN_WORD_PART_LENGTH TYPE OF "dom_WordLength" = 1,
    IN_POSITION_OFFSET TYPE OF "dom_Position" = 0,
    IN_WORD_INDEX_OFFSET TYPE OF "dom_Position" = 0)
RETURNS (
    OUT_WORT_INDEX INTEGER)
AS
DECLARE VARIABLE WORD_INDEX TYPE OF "dom_Position"; /* aktueller Wortindex */
DECLARE VARIABLE TEXT_LEN TYPE OF "dom_Position"; /* Textlnge */
DECLARE VARIABLE TEXT_BUF TYPE OF "dom_TextCache"; /* Puffer fr prc_ftx_index_char-Aufruf */
DECLARE VARIABLE TEXT_BUF_SIZE TYPE OF "dom_WordLength"; /* maximale Gre des Puffers */
DECLARE VARIABLE TEXT_BUF_LENGTH TYPE OF "dom_WordLength"; /* aktuelle Lnge des Puffers */
DECLARE VARIABLE TEXT_COPY_LENGTH "dom_WordLength"; /* in den Puffer zu kopierende Zeichenanzahl */
DECLARE VARIABLE TEXT_POS TYPE OF "dom_Position"; /* aktuelle Position im Text */
DECLARE VARIABLE TEXT_POS_OFFSET TYPE OF "dom_Position"; /* aktueller Offset fr Aufruf von prc_ftx_index_char */
DECLARE VARIABLE LAST_POS TYPE OF "dom_WordLength"; /* Endposition fr Kopie in den Puffer */
DECLARE VARIABLE CHAR_IS_ALPHANUM TYPE OF "dom_BOOL"; /* aktuelles Zeichen ist alphanumerisch oder nicht */
BEGIN

  /* Wordindex initialisieren */
  WORD_INDEX = IN_WORD_INDEX_OFFSET;

  /* Textlnge bestimmen */
  TEXT_LEN = COALESCE(CHAR_LENGTH(IN_TEXT), 0);
  IF (TEXT_LEN > 0)
  THEN 
  BEGIN

    /* Pufferlnge bestimmen */
    EXECUTE PROCEDURE "prc_domain_text_length"('dom_TextCache')
      RETURNING_VALUES TEXT_BUF_SIZE;

    /* erster Textblock */
    TEXT_POS_OFFSET = 0;
    TEXT_POS = 1;
    TEXT_BUF = SUBSTRING(IN_TEXT FROM TEXT_POS FOR MINVALUE(TEXT_LEN, TEXT_BUF_SIZE));
    TEXT_BUF_LENGTH = CHAR_LENGTH(TEXT_BUF);
    TEXT_POS = TEXT_POS + TEXT_BUF_LENGTH;
    /* alle Textblcke abarbeiten */
    WHILE (TEXT_BUF_LENGTH > 0)
    DO
    BEGIN

      /* letzten Wortzwischenraum suchen */
      LAST_POS = TEXT_BUF_LENGTH;
      EXECUTE PROCEDURE "prc_ftx_is_alpha_num"(SUBSTRING(TEXT_BUF FROM LAST_POS FOR 1))
        RETURNING_VALUES CHAR_IS_ALPHANUM;
      WHILE ((LAST_POS > 1) AND (CHAR_IS_ALPHANUM <> 0))
      DO
      BEGIN
        LAST_POS = LAST_POS - 1;
        EXECUTE PROCEDURE "prc_ftx_is_alpha_num"(SUBSTRING(TEXT_BUF FROM LAST_POS FOR 1))
          RETURNING_VALUES CHAR_IS_ALPHANUM;
      END

      /* Textblock indizieren */
      IF (CHAR_IS_ALPHANUM = 0)
      THEN 
      BEGIN
        /* Wortzwischenraum gefunden: bis gefundene Position indizieren */
        EXECUTE PROCEDURE "prc_ftx_index_char"(SUBSTRING(TEXT_BUF FROM 1 FOR LAST_POS - 1), IN_WHOLE_WORD, IN_WORD_BEGINNINGS, IN_WORD_TRAILINGS, IN_WORD_PARTS, IN_MIN_WORD_PART_LENGTH, TEXT_POS_OFFSET + IN_POSITION_OFFSET, WORD_INDEX)
          RETURNING_VALUES WORD_INDEX;
        TEXT_POS_OFFSET = TEXT_POS_OFFSET + LAST_POS;
        TEXT_COPY_LENGTH = TEXT_BUF_LENGTH - LAST_POS;
        IF (TEXT_COPY_LENGTH > 0)
        THEN 
          TEXT_BUF = SUBSTRING(TEXT_BUF FROM LAST_POS + 1 FOR TEXT_COPY_LENGTH);
        ELSE
          TEXT_BUF = '';
      END
      ELSE
      BEGIN
        /* keinen Wortzwischenraum gefunden: ganzen Block als Wort indizieren */
        EXECUTE PROCEDURE "prc_ftx_index_char"(TEXT_BUF, IN_WHOLE_WORD, IN_WORD_BEGINNINGS, IN_WORD_TRAILINGS, IN_WORD_PARTS, IN_MIN_WORD_PART_LENGTH, TEXT_POS_OFFSET + IN_POSITION_OFFSET, WORD_INDEX)
          RETURNING_VALUES WORD_INDEX;
        TEXT_POS_OFFSET = TEXT_POS_OFFSET + TEXT_BUF_LENGTH;
        TEXT_BUF = '';
      END

      /* nchster Block */
      TEXT_COPY_LENGTH = MINVALUE(TEXT_LEN - TEXT_POS + 1, TEXT_BUF_SIZE - CHAR_LENGTH(TEXT_BUF));
      IF (TEXT_COPY_LENGTH > 0)
      THEN 
      BEGIN
        TEXT_BUF = TEXT_BUF || SUBSTRING(IN_TEXT FROM TEXT_POS FOR TEXT_COPY_LENGTH);
        TEXT_POS = TEXT_POS + TEXT_COPY_LENGTH;
      END
      TEXT_BUF_LENGTH = CHAR_LENGTH(TEXT_BUF);

    END /* WHILE (TEXT_POS <= TEXT_LEN) */

  END /* IF (TEXT_LEN > 0) */

  OUT_WORT_INDEX = WORD_INDEX;

END^


CREATE OR ALTER PROCEDURE "prc_ftx_is_alpha_num" (
    IN_CHAR "dom_SingleChar" COLLATE UTF8)
RETURNS (
    OUT_IS_ALPHANUM "dom_BOOL")
AS
BEGIN

  IF (
       (UPPER(IN_CHAR) <> LOWER(IN_CHAR))
       OR
       (IN_CHAR IN ('', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'))
     )
  THEN
  BEGIN
    OUT_IS_ALPHANUM = 1;
  END
  ELSE
  BEGIN
    OUT_IS_ALPHANUM = 0;
  END

/*
  IF (IN_CHAR IN (ASCII_CHAR(9), ASCII_CHAR(10), ASCII_CHAR(11), ASCII_CHAR(13), ASCII_CHAR(32)))
  THEN
  BEGIN
    OUT_IS_ALPHANUM = 0;
  END
  ELSE
  BEGIN
    OUT_IS_ALPHANUM = 1;
  END
*/

END^


CREATE OR ALTER PROCEDURE "prc_ftx_word_index" (
    IN_WORD "dom_Word" COLLATE UTF8,
    IN_WORD_START "dom_Position",
    IN_WORD_INDEX "dom_Position",
    IN_WHOLE_WORD "dom_BOOL",
    IN_WORD_BEGINNINGS "dom_BOOL",
    IN_WORD_TRAILINGS "dom_BOOL",
    IN_WORD_PARTS "dom_BOOL",
    IN_MIN_WORD_PART_LENGTH "dom_WordLength" = 1)
RETURNS (
    OUT_WORD_INDEX "dom_Position")
AS
DECLARE VARIABLE WORD_LENGTH "dom_WordLength"; /* Lnge des Wortes */
DECLARE VARIABLE IDX "dom_WordLength"; /* Laufvariable 1 */
DECLARE VARIABLE IDY "dom_WordLength"; /* Laufvariable 2 */
BEGIN

  OUT_WORD_INDEX = IN_WORD_INDEX;
  WORD_LENGTH = COALESCE(CHAR_LENGTH(IN_WORD), 0);
  IF (WORD_LENGTH > 0)
  THEN
  BEGIN

    /* Wortindex aktualisieren */
    OUT_WORD_INDEX = OUT_WORD_INDEX + 1;

    /* Wort speichern */
    IF (IN_WHOLE_WORD <> 0)
    THEN
      EXECUTE PROCEDURE "prc_ftx_word_link"(IN_WORD, OUT_WORD_INDEX, IN_WORD_START, 0, IN_MIN_WORD_PART_LENGTH);

    /* Wortanfnge speichern */
    IF ((IN_WORD_BEGINNINGS <> 0) AND (WORD_LENGTH > 1))
    THEN
    BEGIN
      IDX = 1;
      WHILE (IDX < WORD_LENGTH)
      DO
      BEGIN
        /* Wortteil speichern */
        EXECUTE PROCEDURE "prc_ftx_word_link"(LEFT(IN_WORD, IDX), OUT_WORD_INDEX, IN_WORD_START, 1, IN_MIN_WORD_PART_LENGTH);
        /* nchster Teil */
        IDX = IDX + 1;
      END
    END

    /* Wortteile ohne Anfang und Ende speichern*/
    IF ((IN_WORD_PARTS <> 0) AND (WORD_LENGTH > 2))
    THEN
    BEGIN
      IDX = 2;
      WHILE (IDX <= WORD_LENGTH - 1)
      DO
      BEGIN
        IDY = IDX + 1;
        WHILE (IDY <= WORD_LENGTH)
        DO
        BEGIN
          /* Wortteil speichern */
          EXECUTE PROCEDURE "prc_ftx_word_link"(SUBSTRING(IN_WORD FROM IDX FOR IDY - IDX), OUT_WORD_INDEX, IN_WORD_START + IDX - 1, 2, IN_MIN_WORD_PART_LENGTH);
          /* nchster Teil */
          IDY = IDY + 1;
        END
        /* nchster Teil */
        IDX = IDX + 1;
      END
    END

    /* Wortenden speichern */
/*  RIGHT funktioniert nicht mit UTF8
    IF ((IN_WORD_TRAILINGS <> 0) AND (WORD_LENGTH > 1))
    THEN
    BEGIN
      IDX = 1;
      WHILE (IDX < WORD_LENGTH)
      DO
      BEGIN
        /-* Wortteil speichern *-/
        EXECUTE PROCEDURE "prc_ftx_word_link"(RIGHT(IN_WORD, IDX), OUT_WORD_INDEX, IN_WORD_START + WORD_LENGTH - IDX, 3);
        /-* nchster Teil *-/
        IDX = IDX + 1;
      END
    END*/
    /* Wortenden speichern */
    IF ((IN_WORD_TRAILINGS <> 0) AND (WORD_LENGTH > 1))
    THEN
    BEGIN
      IDX = 2;
      WHILE (IDX <= WORD_LENGTH)
      DO
      BEGIN
        /* Wortteil speichern */
        EXECUTE PROCEDURE "prc_ftx_word_link"(SUBSTRING(IN_WORD FROM IDX FOR WORD_LENGTH - IDX + 1), OUT_WORD_INDEX, IN_WORD_START + IDX - 1, 3, IN_MIN_WORD_PART_LENGTH);
        /* nchster Teil */
        IDX = IDX + 1;
      END
    END

  END /* IF (WORD_LENGTH > 0) */

END^


CREATE OR ALTER PROCEDURE "prc_ftx_word_link" (
    IN_WORD "dom_Word" COLLATE UTF8,
    IN_WORD_INDEX "dom_Position",
    IN_WORD_POSITION "dom_Position",
    IN_WORD_TYPE "dom_WordType",
    IN_MIN_WORD_PART_LENGTH "dom_WordLength" = 1)
AS
DECLARE VARIABLE WORD_ID TYPE OF "dom_PK"; /* ID des Wortes in Tabelle words */
BEGIN

  /* Wort speichern */
  /*
  MATCHING kommt nicht mit UPPER oder LOWER klar
  UPDATE OR INSERT INTO "words"
    ("CL_wrd_word")
  VALUES
    (:IN_WORD)
  MATCHING
    ("CL_wrd_word")
  RETURNING
    "PK_wrd"
  INTO
    :WORD_ID;
  */

  IF (
       (IN_WORD_TYPE = 0) /* ganzes Wort */
       OR
       (CHAR_LENGTH(IN_WORD) >= IN_MIN_WORD_PART_LENGTH)
     )
  THEN
  BEGIN

    IN_WORD = LOWER(IN_WORD);
    SELECT
      "PK_wrd"
    FROM
      "words"
    WHERE
      LOWER("CL_wrd_word") = :IN_WORD
    INTO
      :WORD_ID;

    IF (WORD_ID IS NULL)
    THEN
      INSERT INTO
        "words"
        ("CL_wrd_word")
      VALUES
        (:IN_WORD)
      RETURNING
        "PK_wrd"
      INTO
        :WORD_ID;

    /* Verknpfung auf das Wort speichern */
    INSERT INTO "_temp_text_link_words"
      (
        "FK_ttw_wrd",
        "CL_ttw_index",
        "CL_ttw_position",
        "FK_ttw_wty"
      )
    VALUES
      (
        :WORD_ID,
        :IN_WORD_INDEX,
        :IN_WORD_POSITION,
        :IN_WORD_TYPE
      );

  END

END^



SET TERM ; ^


/******************************************************************************/
/****                             Descriptions                             ****/
/******************************************************************************/

COMMENT ON DOMAIN "dom_BOOL" IS 
'boolescher Wert';

COMMENT ON DOMAIN "dom_Hash" IS 
'Text-Hash';

COMMENT ON DOMAIN "dom_PK" IS 
'Primrschlssel';

COMMENT ON DOMAIN "dom_Position" IS 
'Position im Text';

COMMENT ON DOMAIN "dom_RDB$NAME" IS 
'Systemname';

COMMENT ON DOMAIN "dom_SingleChar" IS 
'Einzelzeichen';

COMMENT ON DOMAIN "dom_TS" IS 
'Zeitpunkt';

COMMENT ON DOMAIN "dom_Text" IS 
'Textfeld';

COMMENT ON DOMAIN "dom_TextCache" IS 
'Text-Cache';

COMMENT ON DOMAIN "dom_Word" IS 
'Wortfeld';

COMMENT ON DOMAIN "dom_WordLength" IS 
'Wortlnge';



/******************************************************************************/
/****                             Descriptions                             ****/
/******************************************************************************/

COMMENT ON TABLE "_temp_text_link_words" IS 
'Temporre Tabelle fr Indizierung';

COMMENT ON TABLE "text_link_words" IS 
'Text Wort Verknpfung';

COMMENT ON TABLE "texts" IS 
'Texte';

COMMENT ON TABLE "word_types" IS 
'Worttypen';

COMMENT ON TABLE "words" IS 
'Wrter';



/******************************************************************************/
/****                             Descriptions                             ****/
/******************************************************************************/

COMMENT ON PROCEDURE "prc_domain_text_length" IS 
'Textlnge einer Textdomne ermitteln';

COMMENT ON PROCEDURE "prc_ftx_index_char" IS 
'Wrter in einem Text indizieren';

COMMENT ON PROCEDURE "prc_ftx_index_text" IS 
'Wrter in einem Text-BLOB indizieren';

COMMENT ON PROCEDURE "prc_ftx_is_alpha_num" IS 
'Prfung auf Buchstabe oder Ziffer';

COMMENT ON PROCEDURE "prc_ftx_word_index" IS 
'Wort mit Teilen in den Index eintragen';

COMMENT ON PROCEDURE "prc_ftx_word_link" IS 
'Wort mit Worttabelle verknpfen';



/******************************************************************************/
/****                             Descriptions                             ****/
/******************************************************************************/

COMMENT ON GENERATOR "gen_PK" IS 
'Primrschlssel';



/* Fields descriptions */

COMMENT ON COLUMN "_temp_text_link_words"."FK_ttw_wrd" IS 
'Wort';

COMMENT ON COLUMN "_temp_text_link_words"."FK_ttw_wty" IS 
'Typ des Wortes';

COMMENT ON COLUMN "_temp_text_link_words"."CL_ttw_index" IS 
'Index des Wortes im Text';

COMMENT ON COLUMN "_temp_text_link_words"."CL_ttw_position" IS 
'Position des Wortes im Text';

COMMENT ON COLUMN "text_link_words"."FK_tlw_txt" IS 
'FK auf Texte';

COMMENT ON COLUMN "text_link_words"."FK_tlw_wrd" IS 
'FK auf Wrter';

COMMENT ON COLUMN "text_link_words"."FK_tlw_wty" IS 
'Typ des Wortes';

COMMENT ON COLUMN "text_link_words"."CL_tlw_index" IS 
'Index des Wortes im Text';

COMMENT ON COLUMN "text_link_words"."CL_tlw_position" IS 
'Position des Wortes im Text';

COMMENT ON COLUMN "texts"."PK_txt" IS 
'Primrschlssel';

COMMENT ON COLUMN "texts"."CL_txt_text" IS 
'Textfeld';

COMMENT ON COLUMN "texts"."CL_txt_length" IS 
'Textlnge';

COMMENT ON COLUMN "texts"."CL_txt_bytes" IS 
'Textgre';

COMMENT ON COLUMN "texts"."CL_txt_hash" IS 
'Textprfsumme';

COMMENT ON COLUMN "texts"."CL_txt_update_index" IS 
'Flag zum Start der Indizierung';

COMMENT ON COLUMN "texts"."CL_txt_update_delay" IS 
'Flag zum Verzgern der Indizierung';

COMMENT ON COLUMN "texts"."CL_txt_inserted" IS 
'Datensatz angelegt';

COMMENT ON COLUMN "texts"."CL_txt_updated" IS 
'Datensatz gendert';

COMMENT ON COLUMN "word_types"."PK_wty" IS 
'Primrschlssel';

COMMENT ON COLUMN "word_types"."CL_wty_name" IS 
'Typ des Wortes';

COMMENT ON COLUMN "words"."PK_wrd" IS 
'Primrschlssel';

COMMENT ON COLUMN "words"."CL_wrd_word" IS 
'Wort';



/* Parameters descriptions */

COMMENT ON PARAMETER "prc_domain_text_length".IN_DOM_NAME IS 
'Domnenname';

COMMENT ON PARAMETER "prc_domain_text_length".OUT_TEXT_LENGTH IS 
'Textlnge der Domne';

COMMENT ON PARAMETER "prc_ftx_index_char".IN_MIN_WORD_PART_LENGTH IS 
'Mindestlnge des Wortteiles';

COMMENT ON PARAMETER "prc_ftx_index_text".IN_MIN_WORD_PART_LENGTH IS 
'Mindestlnge des Wortteiles';

COMMENT ON PARAMETER "prc_ftx_is_alpha_num".IN_CHAR IS 
'zu prfendes Zeichen';

COMMENT ON PARAMETER "prc_ftx_is_alpha_num".OUT_IS_ALPHANUM IS 
'1: Wort ist alphanumerisch';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_MIN_WORD_PART_LENGTH IS 
'Mindestlnge des Wortteiles';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WHOLE_WORD IS 
'ganzes Wort wird indiziert';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD IS 
'zu indizierendes Wort';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD_BEGINNINGS IS 
'Wortanfnge werden indiziert';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD_INDEX IS 
'Wortindex vorherigen Wortes im Text';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD_PARTS IS 
'Wortteile ohne Wortanfnge und Wortenden werden indiziert';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD_START IS 
'Position des ersten Wortzeichens im Text';

COMMENT ON PARAMETER "prc_ftx_word_index".IN_WORD_TRAILINGS IS 
'Wortenden werden indiziert';

COMMENT ON PARAMETER "prc_ftx_word_index".OUT_WORD_INDEX IS 
'Index des Wortes im Text';

COMMENT ON PARAMETER "prc_ftx_word_link".IN_MIN_WORD_PART_LENGTH IS 
'Mindestlnge des Wortteiles';

COMMENT ON PARAMETER "prc_ftx_word_link".IN_WORD IS 
'zu indizierendes Wort';

COMMENT ON PARAMETER "prc_ftx_word_link".IN_WORD_INDEX IS 
'Index des Wortes im Text';

COMMENT ON PARAMETER "prc_ftx_word_link".IN_WORD_POSITION IS 
'Position des ersten Buchstabens im Text';

COMMENT ON PARAMETER "prc_ftx_word_link".IN_WORD_TYPE IS 
'Typ des Wortes (ganzes Wort, Wortanfang, Wortende oder Wortteil)';



/******************************************************************************/
/****                              Privileges                              ****/
/******************************************************************************/


/* Privileges of users */
GRANT ALL ON RDB$FORMATS TO DBUSER WITH GRANT OPTION;
GRANT ALL ON RDB$PAGES TO DBUSER WITH GRANT OPTION;
GRANT ALL ON RDB$ROLES TO DBUSER WITH GRANT OPTION;
GRANT ALL ON "_temp_text_link_words" TO DBUSER WITH GRANT OPTION;
GRANT ALL ON "text_link_words" TO DBUSER WITH GRANT OPTION;
GRANT ALL ON "texts" TO DBUSER WITH GRANT OPTION;
GRANT ALL ON "word_types" TO DBUSER WITH GRANT OPTION;
GRANT ALL ON "words" TO DBUSER WITH GRANT OPTION;
GRANT EXECUTE ON PROCEDURE "prc_domain_text_length" TO DBUSER;
GRANT EXECUTE ON PROCEDURE "prc_ftx_index_char" TO DBUSER;
GRANT EXECUTE ON PROCEDURE "prc_ftx_index_text" TO DBUSER;
GRANT EXECUTE ON PROCEDURE "prc_ftx_is_alpha_num" TO DBUSER;
GRANT EXECUTE ON PROCEDURE "prc_ftx_word_index" TO DBUSER;
GRANT EXECUTE ON PROCEDURE "prc_ftx_word_link" TO DBUSER;
GRANT SELECT ON RDB$FORMATS TO PUBLIC;
GRANT SELECT ON RDB$PAGES TO PUBLIC;
GRANT SELECT ON RDB$ROLES TO PUBLIC;

/* Privileges of triggers */
GRANT ALL ON "_temp_text_link_words" TO TRIGGER "texts_AIU0";
GRANT SELECT, INSERT, DELETE ON "text_link_words" TO TRIGGER "texts_AIU0";
GRANT SELECT, UPDATE, REFERENCES ON "texts" TO TRIGGER "texts_AIU0";
GRANT EXECUTE ON PROCEDURE "prc_domain_text_length" TO TRIGGER "texts_AIU0";
GRANT EXECUTE ON PROCEDURE "prc_ftx_index_char" TO TRIGGER "texts_AIU0";
GRANT EXECUTE ON PROCEDURE "prc_ftx_index_text" TO TRIGGER "texts_AIU0";
GRANT UPDATE, REFERENCES ON "texts" TO TRIGGER "texts_BIU0";
GRANT UPDATE, REFERENCES ON "words" TO TRIGGER "words_BIU0";

/* Privileges of procedures */
GRANT EXECUTE ON PROCEDURE "prc_ftx_is_alpha_num" TO PROCEDURE "prc_ftx_index_char";
GRANT EXECUTE ON PROCEDURE "prc_ftx_word_index" TO PROCEDURE "prc_ftx_index_char";
GRANT EXECUTE ON PROCEDURE "prc_domain_text_length" TO PROCEDURE "prc_ftx_index_text";
GRANT EXECUTE ON PROCEDURE "prc_ftx_index_char" TO PROCEDURE "prc_ftx_index_text";
GRANT EXECUTE ON PROCEDURE "prc_ftx_is_alpha_num" TO PROCEDURE "prc_ftx_index_text";
GRANT EXECUTE ON PROCEDURE "prc_ftx_word_link" TO PROCEDURE "prc_ftx_word_index";
GRANT ALL ON "_temp_text_link_words" TO PROCEDURE "prc_ftx_word_link";
GRANT SELECT, INSERT ON "words" TO PROCEDURE "prc_ftx_word_link";
